                Často kladené otázky na podporu místního nastavení
                ==================================================

Pokud vám nefunguje podpora I18N, tak zkontrolujte tato nastavení:

        - systémová proměnná LANG musí být nastavená tak, aby odpovídala vašemu
          místnímu nastavení (čeština = czech), vaše místní nastavení musí být
          podporovatelné vaším operačním systémem nebo X emulací. Všechny
          podporovatelné lokalizace zobrazíte příkazem "locale -a". Měli byste
          také zkontrolovat, zda-li je vaše místní nastavení podporované
          X emulací. Soubor: "/usr/X11R6/lib/X11/locale/locale.alias" by měl
          obsahovat podobnou sekvenci (podpora pro češtinu):

                cs                              cs_CZ.ISO8859-2
                cs_CS                           cs_CZ.ISO8859-2
                cs_CS.ISO8859-2                 cs_CZ.ISO8859-2
                cs_CZ                           cs_CZ.ISO8859-2
                cs_CZ.iso88592                  cs_CZ.ISO8859-2
                cz                              cz_CZ.ISO8859-2
                cz_CZ                           cz_CZ.ISO8859-2

        - zkontrolujte použití správných fontů pro vaše místní nastavení.
          Pokud nepoužíváte správné fonty s kódováním, které je nastaveno v
          Xlib nebo libc, tak se můžou dít "dost divné věci". Zkuste explicitně
          zadat kování do proměnné LANG: LANG=cs_CS.ISO8859-2 nebo
          LANG=cs_CZ.iso88592 a znovu zkontrolujte:
          "/usr/X11R6/lib/X11/locale/locale.alias"

        - pokud váš operační systém nepodporuje místní nastavení (locales), nebo
          pokud váš OS nepodporuje místní nastavení pro váš jazyk, můžete
          použít emulaci místního nastavení X Window Systému. Zprovoznit tuto
          emulaci je možné spuštěním ./configure s volbou "--witn-x-locale".
          Pokud používáte nějaký komerční systém jako je například IRIX, AIX,
          Solaris, ...,tak asi X emulaci nebudete potřebovat. Ovšem pokud je
          vaším operačním systémem Linux, NetBSD nebo jiný užasný, volně
          šiřitelný operační systém, tak je možné, že vaše místní nastavení
          zatím není podporované. Potom použijte volbu "--witn-x-locale".

          Pozn: Aby jste mohli používat X emulaci místních nastavení, tak vaše
                Xlib musí být zkompilované s touto podporou. Xlib v RedHat 5.0
                tak zkompilované nejsou (RH4.x jsou OK). Překompilované Xlib
                s podporou pro emulací místních nastavení pro RH5.0 jsou
                dostupné na adrese:

                ftp://ftp.linux.or.jp/pub/RPM/glibc

        - fonty, které používáte by měli být podporované vaším místním nastavením.
          Jestliže vaše nastavení fontů v souboru ~/GNUstep/Defaults/WindowMaker
          vypadá takto:

          WindowTitleFont = "-*-helvetica-bold-r-normal-*-12-*-*-*-*-*-*-*";
          MenuTitleFont = "-*-helvetica-bold-r-normal-*-12-*-*-*-*-*-*-*";  ,

          nemůžete zobrazovat znaky jiné než iso8859-x ve fontu helvetica.
          Jednoduchý způsob, jak zobrazovat znaky různých jazyků, je nastavit všechny
          fonty takto:

                "-*-*-medium-r-normal-*-14-*-*-*-*-*-*-*"

          Také je nutné změnit nastavení fontů v souborech stylů v adresáři:
          ~/Library/WindowMaker/Style.

        - pokud si nejste jisti, zda-li má systémová proměnná LC_TYPE správnou
          hodnotu, tak ji nenastavujte.